# -*- coding: UTF-8 -*-

import os.path
import time
from typing import Optional, Union, Literal

import numpy as np

import pandas as pd
from anndata import AnnData
from pandas import DataFrame

from .. import util as ul
from ..tool import RandomWalk, overlap_sum, obtain_cell_cell_network, calculate_init_score_weight

from ..file import save_h5ad, save_pkl, read_h5ad, read_pkl
from ..preprocessing import filter_data, poisson_vi
from ..util import path, enrichment_optional, to_dense, collection, to_sparse, difference_peak_optional, project_name

__name__: str = "model_core"


def _run_random_walk_(random_walk: RandomWalk, is_ablation: bool, is_simple: bool) -> AnnData:
    start_time = time.perf_counter()

    if not random_walk.is_run_core:
        random_walk.run_core()

    if not random_walk.is_run_enrichment:
        random_walk.run_enrichment()

    if is_ablation and not is_simple:
        if not random_walk.is_benchmark:
            random_walk.run_benchmark()

        if not random_walk.is_run_ablation_ncw:
            random_walk.run_ablation_ncw()

        if not random_walk.is_run_ablation_nsw:
            random_walk.run_ablation_nsw()

        if not random_walk.is_run_ablation_ncsw:
            random_walk.run_ablation_ncsw()

        if not random_walk.is_run_ablation_m_knn:
            random_walk.run_ablation_m_knn()

        if not random_walk.is_run_en_ablation_ncw:
            random_walk.run_en_ablation_ncw()

        if not random_walk.is_run_en_ablation_nsw:
            random_walk.run_en_ablation_nsw()

        if not random_walk.is_run_en_ablation_ncsw:
            random_walk.run_en_ablation_ncsw()

        if not random_walk.is_run_en_ablation_m_knn:
            random_walk.run_en_ablation_m_knn()

    random_walk.elapsed_time += time.perf_counter() - start_time

    return random_walk.trs_adata


def core(
    adata: AnnData,
    variants: dict,
    trait_info: DataFrame,
    cell_rate: Optional[float] = None,
    peak_rate: Optional[float] = None,
    max_epochs: int = 500,
    lr: float = 1e-4,
    batch_size: int = 128,
    eps: float = 1e-08,
    early_stopping: bool = True,
    early_stopping_patience: int = 50,
    batch_key: Optional[str] = None,
    resolution: float = 0.5,
    k: int = 50,
    or_k: int = 25,
    weight: float = 0.1,
    kernel: Literal["laplacian", "gaussian"] = "gaussian",
    local_k: int = 10,
    kernel_gamma: Optional[Union[float, collection]] = None,
    epsilon: float = 1e-05,
    gamma: float = 0.05,
    enrichment_gamma: float = 0.05,
    p: int = 2,
    n_jobs: int = -1,
    min_seed_cell_rate: float = 0.01,
    max_seed_cell_rate: float = 0.05,
    credible_threshold: float = 0,
    diff_peak_value: difference_peak_optional = 'emp_effect',
    enrichment_threshold: Union[enrichment_optional, float] = 'golden',
    is_ablation: bool = False,
    model_dir: Optional[path] = None,
    save_path: Optional[path] = None,
    is_simple: bool = True,
    is_save_random_walk_model: bool = False,
    is_file_exist_loading: bool = False,
    filename_dict: Optional[dict] = None,
    block_size: int = -1
) -> AnnData:
    """
    The core algorithm of sciv includes the flow of all algorithms, as well as drawing and saving data.
    In the entire algorithm, the samples are in the row position, and the traits or diseases are in the column position,
        while ensuring that there is no interaction between the traits or diseases,
        ensuring the stability of the results;
    Meaning of main variables:
        1. `overlap_adata`, (obs: peaks, var: traits/diseases) Peaks-traits/diseases data obtained by overlaying variant
         data with peaks.
        2. `da_peaks`, (obs: clusters (Leiden), var: peaks) Differential peak data of cell clustering, used for weight
         correction of cells.
        3. `init_score`, (obs: cells, var: traits/diseases) This is the initial TRS data.
        4. `cc_data`, (obs: cells, var: cells) Cell similarity data.
        5. `random_walk`, RandomWalk class.
        6. `trs`, (obs: cells, var: traits/diseases) This is the final TRS data.
    :param adata: scATAC-seq data;
    :param variants: variant data; This data is recommended to be obtained by executing the `fl.read_variants` method.
    :param trait_info: variant annotation file information;
    :param cell_rate: Removing the percentage of cell count in total cell count only takes effect when the min_cells
        parameter is None;
    :param peak_rate: Removing the percentage of peak count in total peak count only takes effect when the min_peaks
        parameter is None;
    :param max_epochs: The maximum number of epochs for PoissonVI training;
    :param lr: Learning rate for optimization;
    :param batch_size: Minibatch size to use during training;
    :param eps: Optimizer eps;
    :param early_stopping: Whether to perform early stopping with respect to the validation set;
    :param early_stopping_patience: How many epochs to wait for improvement before early stopping;
    :param batch_key: Batch information in scATAC-seq data;
    :param resolution: Resolution of the Leiden Cluster. The recommended values are any one of 0.4, 0.9, 1.3, 1.5;
    :param k: When building an mKNN network, the number of nodes connected by each node (and operation);
    :param or_k: When building an mKNN network, the number of nodes connected by each node (or operation);
    :param weight: The weight of interactions or operations;
    :param local_k: Determining the number of neighbors for the adaptive kernel;
    :param kernel: Determine the kernel function to be used;
    :param kernel_gamma: If None, it defaults to the adaptive value obtained through the local information of
        parameter `local_k`. Otherwise, it should be strictly positive;
    :param epsilon: conditions for stopping in random walk;
    :param gamma: reset weight for random walk;
    :param enrichment_gamma: reset weight for random walk for enrichment;
    :param p: Distance used for loss {1: Manhattan distance, 2: Euclidean distance};
    :param n_jobs: The maximum number of concurrently running jobs;
    :param min_seed_cell_rate: The minimum percentage of seed cells in all cells;
    :param max_seed_cell_rate: The maximum percentage of seed cells in all cells;
    :param credible_threshold: The threshold for determining the credibility of enriched cells in the context of
        enrichment, i.e. the threshold for judging enriched cells;
    :param diff_peak_value: Specify the correction value in peak correction of clustering type differences.
        {'emp_effect', 'bayes_factor', 'emp_prob1'}
    :param enrichment_threshold: Only by setting a threshold for the standardized output TRS can a portion of the
        enrichment results be obtained. Parameters support string types {'golden', 'half', 'e', 'pi', 'none'},
        or valid floating-point types within the range of (0, log1p(1)).
    :param is_ablation: True represents obtaining the results of the ablation experiment. This parameter is limited by
        the `is_simple` parameter, and its effectiveness requires setting `is_simple` to `False`;
    :param model_dir: The folder name saved by the training module;
        It is worth noting that if the training model file (`model.pt`) exists in this path, it will be automatically
        read and skip the training of `PoissonVI` model.
    :param save_path: Save path for process files and result files;
    :param is_simple: True represents not adding unnecessary intermediate variables, only adding the final result.
        It is worth noting that when set to `True`, the `is_ablation` parameter will become invalid, and when set to
        `False`, `is_ablation` will only take effect;
    :param is_save_random_walk_model: Default to `False`, do not save random walk model. When setting `True`, please
        ensure sufficient storage as the saved `pkl` file is relatively large.
    :param is_file_exist_loading: By default, the file will be overwritten. When set to `True`, if the file exists, the
        process will be skipped and the file will be directly read as the result;
    :param filename_dict: The name of the file that exists.
        default: {
            "sc_atac": "sc_atac.h5ad",
            "da_peaks": "da_peaks.h5ad",
            "atac_overlap": "atac_overlap.h5ad",
            "init_score": "init_score.h5ad",
            "cc_data": "cc_data.h5ad",
            "random_walk": "random_walk.h5ad",
            "trs": "trs.h5ad"
        }
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        By sacrificing time and space to reduce memory consumption to a certain extent.
        If the value is less than or equal to zero, no block operation will be performed.
    :return: `trs`, (obs: cells, var: traits/diseases) This is the final TRS data.
    """

    # start time
    start_time = time.perf_counter()

    if len(variants.keys()) == 0:
        ul.log(__name__).error("The number of mutations is empty.")
        raise ValueError("The number of mutations is empty.")

    _trait_count_ = trait_info.shape[0]

    if len(variants.keys()) != _trait_count_:
        ul.log(__name__).error(
            "The parameters `variants` and `trait_info` are inconsistent. "
            "These two parameters can be obtained using method `fl.read_variants`."
        )
        raise ValueError(
            "The parameters `variants` and `trait_info` are inconsistent. "
            "These two parameters can be obtained using method `fl.read_variants`."
        )

    if adata.shape[0] == 0:
        ul.log(__name__).error("The scATAC-seq data is empty.")
        raise ValueError("The scATAC-seq data is empty.")

    peak_columns: list = list(adata.var.columns)

    if "chr" not in peak_columns or "start" not in peak_columns or "end" not in peak_columns:
        ul.log(__name__).error(
            f"The peaks information {peak_columns} in data `adata` must include three columns: "
            f"`chr`, `start` and `end`. (It is recommended to use the `fl.read_sc_atac` method.)"
        )
        raise ValueError(
            f"The peaks information {peak_columns} in data `adata` must include three columns: "
            f"`chr`, `start` and `end`. (It is recommended to use the `fl.read_sc_atac` method.)"
        )

    if batch_key is not None and batch_key not in adata.obs.columns:
        ul.log(__name__).error(
            f"The cells information {adata.obs.columns} in data `adata` must include the {batch_key} column."
        )
        raise ValueError(
            f"The cells information {adata.obs.columns} in data `adata` must include the {batch_key} column."
        )

    if cell_rate is not None:

        if cell_rate <= 0 or cell_rate >= 1:
            ul.log(__name__).error("The parameter of `cell_rate` should be between 0 and 1.")
            raise ValueError("The parameter of `cell_rate` should be between 0 and 1.")

    if peak_rate is not None:

        if peak_rate <= 0 or peak_rate >= 1:
            ul.log(__name__).error("The parameter of `peak_rate` should be between 0 and 1.")
            raise ValueError("The parameter of `peak_rate` should be between 0 and 1.")

    if resolution <= 0:
        ul.log(__name__).error("The parameter `resolution` must be greater than zero.")
        raise ValueError("The parameter `resolution` must be greater than zero.")

    if k <= 0:
        ul.log(__name__).error("The `k` parameter must be a natural number greater than 0.")
        raise ValueError("The `k` parameter must be a natural number greater than 0.")

    if or_k <= 0:
        ul.log(__name__).error("The `or_k` parameter must be a natural number greater than 0.")
        raise ValueError("The `or_k` parameter must be a natural number greater than 0.")

    if k < or_k:
        ul.log(__name__).warning(
            "The parameter value of `or_k` is greater than the parameter value of `k`, "
            "which is highly likely to result in poor performance."
        )

    if local_k <= 0:
        ul.log(__name__).error("The `local_k` parameter must be a natural number greater than 0.")
        raise ValueError("The `local_k` parameter must be a natural number greater than 0.")

    if kernel not in ["laplacian", "gaussian"]:
        ul.log(__name__).error("Parameter `kernel` only supports two values, `laplacian` and `gaussian`.")
        raise ValueError("Parameter `kernel` only supports two values, `laplacian` and `gaussian`.")

    if weight < 0 or weight > 1:
        ul.log(__name__).error("The parameter of `weight` should be between 0 and 1.")
        raise ValueError("The parameter of `weight` should be between 0 and 1.")

    if gamma < 0 or gamma > 1:
        ul.log(__name__).error("The parameter of `gamma` should be between 0 and 1.")
        raise ValueError("The parameter of `gamma` should be between 0 and 1.")

    if enrichment_gamma < 0 or enrichment_gamma > 1:
        ul.log(__name__).error("The parameter of `enrichment_gamma` should be between 0 and 1.")
        raise ValueError("The parameter of `enrichment_gamma` should be between 0 and 1.")

    if min_seed_cell_rate < 0 or min_seed_cell_rate > 1:
        ul.log(__name__).error("The parameter of `min_seed_cell_rate` should be between 0 and 1.")
        raise ValueError("The parameter of `min_seed_cell_rate` should be between 0 and 1.")

    if max_seed_cell_rate < 0 or max_seed_cell_rate > 1:
        ul.log(__name__).error("The parameter of `max_seed_cell_rate` should be between 0 and 1.")
        raise ValueError("The parameter of `max_seed_cell_rate` should be between 0 and 1.")

    if epsilon > 0.1:
        ul.log(__name__).warning(
            f"Excessive value of parameter `epsilon`=({epsilon}) can lead to "
            f"incorrect iteration and poor enrichment effect."
        )
    elif epsilon <= 0:
        ul.log(__name__).error("The parameter of `epsilon` must be greater than zero.")
        raise ValueError("The parameter of `epsilon` must be greater than zero.")

    if p <= 0:
        ul.log(__name__).error("The `p` parameter must be a natural number greater than 0.")
        raise ValueError("The `p` parameter must be a natural number greater than 0.")
    elif p > 3:
        ul.log(__name__).warning("Suggested value for `p` is 1 or 2.")

    if isinstance(enrichment_threshold, float):

        if enrichment_threshold <= 0 or enrichment_threshold >= np.log1p(1):
            ul.log(__name__).warning(
                "The `enrichment_threshold` parameter is not set within the range of (0, log1p(1)), this parameter "
                "will become invalid."
            )
            ul.log(__name__).warning(
                "It is recommended to set the `enrichment_threshold` parameter to the 'golden' value."
            )

    elif enrichment_threshold not in ["golden", "half", "e", "pi", "none"]:
        ul.log(__name__).error(
            "Invalid enrichment settings. The string type in the `enrichment_threshold` parameter only supports the "
            "following parameter 'golden', 'half', 'e', 'pi', 'none',  Alternatively, input a floating-point type "
            "value within the range of (0, log1p(1))"
        )
        raise ValueError(
            "Invalid enrichment settings. The string type in the `enrichment_threshold` parameter only supports the "
            "following parameter 'golden', 'half', 'e', 'pi', 'none',  Alternatively, input a floating-point type "
            "value within the range of (0, log1p(1))"
        )

    if diff_peak_value not in ['emp_effect', 'bayes_factor', 'emp_prob1', 'all']:
        ul.log(__name__).error(
            "The `diff_peak_value` parameter only supports one of the "
            "{'emp_effect', 'bayes_factor', 'emp_prob1', 'all'} values."
        )
        raise ValueError(
            "The `diff_peak_value` parameter only supports one of the "
            "{'emp_effect', 'bayes_factor', 'emp_prob1', 'all'} values."
        )

    # parameter information
    params: dict = {
        "cell_rate": cell_rate,
        "peak_rate": peak_rate,
        "max_epochs": int(max_epochs),
        "lr": lr,
        "batch_size": batch_size,
        "eps": eps,
        "early_stopping": early_stopping,
        "early_stopping_patience": early_stopping_patience,
        "batch_key": batch_key,
        "resolution": resolution,
        "k": k,
        "or_k": or_k,
        "weight": weight,
        "kernel": kernel,
        "local_k": local_k,
        "kernel_gamma": kernel_gamma,
        "epsilon": epsilon,
        "gamma": gamma,
        "enrichment_gamma": enrichment_gamma,
        "p": p,
        "n_jobs": n_jobs,
        "min_seed_cell_rate": min_seed_cell_rate,
        "max_seed_cell_rate": max_seed_cell_rate,
        "credible_threshold": credible_threshold,
        "enrichment_threshold": enrichment_threshold,
        "diff_peak_value": diff_peak_value,
        "is_ablation": is_ablation,
        "model_dir": str(model_dir),
        "save_path": str(save_path),
        "is_simple": is_simple,
        "is_save_random_walk_model": is_save_random_walk_model,
        "is_file_exist_loading": is_file_exist_loading,
        "filename_dict": filename_dict,
        "block_size": block_size
    }

    """
    Save information
    """

    if save_path is not None:
        save_path = str(save_path)
        ul.file_method(__name__).makedirs(save_path)
    else:
        is_file_exist_loading = False

    if filename_dict is None:
        filename_dict = {
            "sc_atac": "sc_atac.h5ad",
            "da_peaks": "da_peaks.h5ad",
            "atac_overlap": "atac_overlap.h5ad",
            "init_score": "init_score.h5ad",
            "cc_data": "cc_data.h5ad",
            "random_walk": "random_walk.h5ad",
            "trs": "trs.h5ad"
        }

    def _get_file_(_id_: str) -> str:
        return os.path.join(save_path, str(filename_dict.get(_id_, f"{_id_}.h5ad"))) if save_path else None

    # Assign a name to the formed document
    adata_save_file = _get_file_("sc_atac")
    da_peaks_save_file = _get_file_("da_peaks")
    atac_overlap_save_file = _get_file_("atac_overlap")
    init_score_save_file = _get_file_("init_score")
    cc_data_save_file = _get_file_("cc_data")
    random_walk_save_file = _get_file_("random_walk")
    trs_save_file = _get_file_("trs")

    """
    1. Filter scATAC-seq data, PoissonVI
    """

    adata_is_read: bool = False
    da_peaks_is_read: bool = False

    if is_file_exist_loading:

        if os.path.exists(adata_save_file):
            adata = read_h5ad(adata_save_file)
            adata_is_read = True

            if "step" not in adata.uns.keys():
                filter_data(adata, cell_rate=cell_rate, peak_rate=peak_rate)
                adata_is_read = False
        else:
            filter_data(adata, cell_rate=cell_rate, peak_rate=peak_rate)

        if os.path.exists(da_peaks_save_file) and adata.uns["step"] == 1:
            da_peaks = read_h5ad(da_peaks_save_file)
            da_peaks_is_read = True
        else:
            # PoissonVI
            da_peaks = poisson_vi(
                adata,
                max_epochs=max_epochs,
                lr=lr,
                batch_size=batch_size,
                eps=eps,
                early_stopping=early_stopping,
                early_stopping_patience=early_stopping_patience,
                batch_key=batch_key,
                resolution=resolution,
                model_dir=model_dir
            )

    else:
        filter_data(adata, cell_rate=cell_rate, peak_rate=peak_rate)

        da_peaks = poisson_vi(
            adata,
            max_epochs=max_epochs,
            lr=lr,
            batch_size=batch_size,
            eps=eps,
            early_stopping=early_stopping,
            early_stopping_patience=early_stopping_patience,
            batch_key=batch_key,
            resolution=resolution,
            model_dir=model_dir
        )

    poisson_vi_time = adata.uns["elapsed_time"] + da_peaks.uns["elapsed_time"]

    if save_path is not None:

        if not adata_is_read:
            save_h5ad(adata, file=adata_save_file)

        if not da_peaks_is_read:
            save_h5ad(da_peaks, file=da_peaks_save_file)

    """
    2. Overlap regional data and mutation data and sum the PP values of all mutations
       in a region as the values for that region
    """

    # Determine whether it is necessary to read the file
    overlap_is_read: bool = is_file_exist_loading and os.path.exists(atac_overlap_save_file)

    if overlap_is_read:
        overlap_adata: AnnData = read_h5ad(atac_overlap_save_file)

        if overlap_adata.var.shape[0] != _trait_count_:
            ul.log(__name__).warning(
                f"The number of diseases read from file `atac_overlap.h5ad` are inconsistent with the input ({overlap_adata.var.shape[0]} != {_trait_count_}). "
                f"Please check and verify. If the verification is not as expected, file `atac_overlap.h5ad` needs to be moved or deleted."
            )

    else:
        overlap_adata: AnnData = overlap_sum(adata, variants, trait_info, n_jobs=n_jobs)

    del variants, trait_info

    overlap_time = overlap_adata.uns["elapsed_time"]

    if save_path is not None and not overlap_is_read:
        save_h5ad(overlap_adata, file=atac_overlap_save_file)

    """
    3. Calculate the initial trait relevance scores for each cell
    """

    init_score_is_read: bool = is_file_exist_loading and os.path.exists(init_score_save_file)

    if init_score_is_read:
        init_score: AnnData = read_h5ad(init_score_save_file)

        if init_score.var.shape[0] != _trait_count_:
            ul.log(__name__).warning(
                f"The number of diseases read from file `init_score.h5ad` are inconsistent with the input ({init_score.var.shape[0]} != {_trait_count_}). "
                f"Please check and verify. If the verification is not as expected, file `init_score.h5ad` needs to be moved or deleted."
            )

    else:
        # intermediate score data, integration data
        init_score: AnnData = calculate_init_score_weight(
            adata=adata,
            da_peaks_adata=da_peaks,
            overlap_adata=overlap_adata,
            diff_peak_value=diff_peak_value,
            is_simple=is_simple,
            block_size=block_size
        )

    del da_peaks, overlap_adata

    init_score_time = init_score.uns["elapsed_time"]

    if save_path is not None and not init_score_is_read:
        save_h5ad(init_score, file=init_score_save_file)

    """
    4. Calculate cell-cell correlation. Building a network between cells.
    """

    cc_data_is_read: bool = is_file_exist_loading and os.path.exists(cc_data_save_file)

    if cc_data_is_read:
        cc_data: AnnData = read_h5ad(cc_data_save_file)
    else:
        # cell-cell network
        cc_data = obtain_cell_cell_network(
            adata=adata,
            k=k,
            or_k=or_k,
            weight=weight,
            kernel=kernel,
            local_k=local_k,
            gamma=kernel_gamma,
            is_simple=is_simple
        )

    del adata

    smknn_time = cc_data.uns["elapsed_time"]

    if save_path is not None and not cc_data_is_read:
        save_h5ad(cc_data, file=cc_data_save_file)

    """
    5. Random walk with weighted seed cells
    """

    random_walk_is_read: bool = is_file_exist_loading and os.path.exists(random_walk_save_file) and is_save_random_walk_model

    if random_walk_is_read:
        random_walk: RandomWalk = read_pkl(random_walk_save_file)
    else:
        # random walk
        # noinspection DuplicatedCode
        random_walk: RandomWalk = RandomWalk(
            cc_adata=cc_data,
            init_status=init_score,
            epsilon=epsilon,
            gamma=gamma,
            enrichment_gamma=enrichment_gamma,
            p=p,
            n_jobs=n_jobs,
            min_seed_cell_rate=min_seed_cell_rate,
            max_seed_cell_rate=max_seed_cell_rate,
            credible_threshold=credible_threshold,
            enrichment_threshold=enrichment_threshold,
            is_ablation=is_ablation,
            is_simple=is_simple
        )

        if save_path is not None and random_walk_is_read:
            save_pkl(random_walk, save_file=random_walk_save_file)

    del random_walk_is_read, init_score, cc_data

    trs = _run_random_walk_(random_walk, is_ablation, is_simple)
    trs.uns["params"] = params

    del params

    random_walk_time = random_walk.elapsed_time

    # end time
    elapsed_time = time.perf_counter() - start_time
    step_time = poisson_vi_time + overlap_time + init_score_time + smknn_time + random_walk_time

    if elapsed_time < step_time:
        elapsed_time = step_time

    ul.log(__name__).info(f"Algorithm {project_name} consumes a total of {elapsed_time:.2f}s.")

    trs.uns["elapsed_time"] = {
        "PoissonVI": poisson_vi_time,
        "Overlap": overlap_time,
        "initial TRS": init_score_time,
        "SM-kNN": smknn_time,
        "Random walk": random_walk_time,
        "Total time": elapsed_time
    }

    if save_path is not None:
        save_h5ad(trs, file=trs_save_file)

    return trs


def association_score(
    adata: AnnData,
    score_name: str = "association_score",
    layer: str = "trs_source",
    axis: Literal[0, 1] = 0
) -> None:
    trs_source = to_dense(adata.layers[layer], is_array=True)

    trs_source_copy = trs_source.copy()
    trs_source_copy[trs_source_copy > 0] = 1

    trs_count_sum = trs_source_copy.sum(axis=axis)
    trs_count_sum[trs_count_sum == 0] = 1
    relevance_value = trs_source.sum(axis=axis) / trs_count_sum

    if axis == 0:
        adata.var[score_name] = np.array(relevance_value).flatten() * adata.shape[1]
        ul.log(__name__).info(f"View this related result `adata.var[{score_name}]`.")
    elif axis == 1:
        adata.obs[score_name] = np.array(relevance_value).flatten() * adata.shape[0]
        ul.log(__name__).info(f"View this related result `adata.obs[{score_name}]`.")
    else:
        ul.log(__name__).error("The `axis` parameter supports only 0 and 1.")
        raise ValueError("The `axis` parameter supports only 0 and 1.")

    sorted_indices = np.argsort(relevance_value)[::,]
    ranks = np.zeros_like(relevance_value).tolist()

    for rank, index in enumerate(sorted_indices, start=1):
        ranks[index] = rank

    adata.var["rank"] = ranks
    adata.uns["association_score"] = {
        "layer": layer,
        "axis": axis,
        "causal_variant_sets_count": len(ranks)
    }


def knock(
    trs: AnnData,
    sc_atac: AnnData,
    da_peaks: AnnData,
    cc_data: AnnData,
    knock_trait: str,
    knock_info: dict[str, Union[str, collection]],
    knock_value: float = 0,
    is_add_control: bool = False
) -> AnnData:

    if "params" not in trs.uns:
        ul.log(__name__).error("`params` is not in `trs.uns`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")
        raise ValueError("`params` is not in `trs.uns`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")

    if "variants" not in trs.uns:
        ul.log(__name__).error("`variants` is not in `trs.uns`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")
        raise ValueError("`variants` is not in `trs.uns`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")

    if "trait_info" not in trs.uns:
        ul.log(__name__).error("`trait_info` is not in `trs.uns`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")
        raise ValueError("`trait_info` is not in `trs.uns`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")

    if "trs_source" not in trs.layers:
        ul.log(__name__).error("`trs_source` is not in `trs.layers`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")
        raise ValueError("`trs_source` is not in `trs.layers`, please execute function `ml.core` first to obtain the result as input for the `trs` parameter.")

    if knock_value >= 1E-3:
        ul.log(__name__).warning("The value set for `knock_value` here is greater than 1e-3, which can easily fail to achieve the desired effect.")

    # param information
    params = trs.uns["params"]

    # variant information
    variants = trs.uns["variants"]
    trait_info = trs.uns["trait_info"]

    knock_variants: dict = {}
    knock_variants_bg: dict = {}

    if knock_trait not in trs.var["id"]:
        ul.log(__name__).error(f"`{knock_trait}` trait or disease does not exist.")
        raise ValueError(f"`{knock_trait}` trait or disease does not exist.")

    trait_obs = variants[knock_trait].obs
    knock_variant: AnnData = variants[knock_trait].copy()
    knock_variant_value = to_dense(knock_variant.X)

    _trait_info_ = trait_info[trait_info["id"] == knock_trait].values[0]
    knock_trait_info = pd.DataFrame(columns=trait_info.columns)

    ul.log(__name__).info(f"Knockdown or knockout settings for `{knock_trait}` trait or disease.")

    for k, v in knock_info.items():

        if isinstance(v, str):
            if v not in trait_obs["rsId"].tolist():
                ul.log(__name__).error(f"`{v}` variant does not exist in `{knock_trait}` trait or disease")
                raise ValueError(f"`{v}` variant does not exist in `{knock_trait}` trait or disease")
        elif isinstance(v, collection):
            if not set(v).issubset(set(trait_obs["rsId"].tolist())):
                ul.log(__name__).error(f"`{v}` variants does not exist in `{knock_trait}` trait or disease")
                raise ValueError(f"`{v}` variants does not exist in `{knock_trait}` trait or disease")

        need_k_index = trait_obs["rsId"].isin(list([v] if isinstance(v, str) else v)).values
        knock_variant_k = knock_variant.copy()
        knock_variant_k_value = knock_variant_value.copy()
        knock_variant_k_value[need_k_index, :] = knock_value
        knock_variant_k.X = to_sparse(knock_variant_k_value)
        knock_variants[k] = knock_variant_k

        knock_variant_k_bg = knock_variant.copy()
        knock_variant_k_value_bg = knock_variant_value.copy()
        knock_variant_k_value_bg[~need_k_index, :] = knock_value
        knock_variant_k_bg.X = to_sparse(knock_variant_k_value_bg)
        knock_variants_bg[k] = knock_variant_k_bg

        _trait_info_[0] = k
        knock_trait_info.loc[knock_trait_info.shape[0], :] = _trait_info_

    knock_trait_info.index = knock_trait_info["id"].astype(str)

    ul.log(__name__).info(f"Run the process after knocking down or knocking out.")
    knock_overlap_adata: AnnData = overlap_sum(sc_atac, knock_variants, knock_trait_info)

    # intermediate score data, integration data
    knock_init_score: AnnData = calculate_init_score_weight(
        adata=sc_atac,
        da_peaks_adata=da_peaks,
        overlap_adata=knock_overlap_adata,
        diff_peak_value=params["diff_peak_value"],
        is_simple=True,
        block_size=params["block_size"]
    )

    # random walk
    # noinspection DuplicatedCode
    knock_random_walk: RandomWalk = RandomWalk(
        cc_adata=cc_data,
        init_status=knock_init_score,
        epsilon=params["epsilon"],
        gamma=params["gamma"],
        enrichment_gamma=params["enrichment_gamma"],
        p=params["p"],
        n_jobs=params["n_jobs"] if "n_jobs" in params else -1,
        min_seed_cell_rate=params["min_seed_cell_rate"],
        max_seed_cell_rate=params["max_seed_cell_rate"],
        credible_threshold=params["credible_threshold"],
        enrichment_threshold=params["enrichment_threshold"],
        is_ablation=False,
        is_simple=True
    )

    # RUN
    knock_random_walk.run_knock(trs, knock_trait, False)

    if is_add_control:
        knock_random_walk.run_knock(trs, knock_trait, True)

    # Obtain result data
    knock_trs = knock_random_walk.trs_adata

    knock_trs.uns["params"] = {
        "knock_trait": knock_trait,
        "knock_info": knock_info,
        "knock_value": knock_value,
        "is_add_control": is_add_control,
        "run_core_params": params
    }

    trs.uns["is_knock"] = True

    if "pp_sum" in knock_trs.var.columns:
        knock_trs.var["pp_sum"] = knock_trs.var["pp_sum"].astype(float)

    if "pp_mean" in knock_trs.var.columns:
        knock_trs.var["pp_mean"] = knock_trs.var["pp_mean"].astype(float)

    if "count" in knock_trs.var.columns:
        knock_trs.var["count"] = knock_trs.var["count"].astype(int)

    # save result
    if params["save_path"] is not None:
        save_h5ad(knock_trs, file=os.path.join(params["save_path"], f"knock_trs_{knock_trait}.h5ad"))

    return knock_trs



# -*- coding: UTF-8 -*-

from typing import Optional, Tuple, Union, Any

from pandas import DataFrame
from anndata import AnnData
from matplotlib import pyplot as plt
import matplotlib.dates as mdates
import seaborn as sns

from .. import util as ul
from ..preprocessing import adata_map_df
from ..util import path, plot_color_types, collection, plot_end, plot_start

__name__: str = "plot_line"


def stability_line(
    data: Union[AnnData, DataFrame],
    x: str,
    y: str,
    layer: Optional[str] = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    title: Optional[str] = None,
    x_name: Optional[str] = None,
    y_name: Optional[str] = None,
    label: Optional[str] = None,
    legend: Optional[str] = None,
    legend_list: list = None,
    start_color_index: int = 0,
    color_step_size: int = 0,
    color_type: str = "set",
    colors: list = None,
    line_width: float = 1.5,
    x_name_rotation: float = 65,
    x_ticks: Optional[Union[int, collection]] = None,
    y_limit: Tuple[float, float] = (0, 1),
    output: Optional[path] = None,
    is_str: bool = True,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:

    fig, ax = plot_start(width, height, bottom, output, show)

    new_data = data.copy()

    if isinstance(new_data, AnnData):

        if legend_list is not None:

            index_list = []
            label_list = list(new_data.var[label])

            for lab in range(len(label_list)):
                if legend_list.count(label_list[lab]) > 0:
                    index_list.append(lab)

            if legend_list is not None:
                new_data = new_data[:, index_list]

        # judge layers
        if layer is not None:

            if layer not in list(new_data.layers):
                ul.log(__name__).error("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
                raise ValueError("The value of the `layer` parameter must be one of the keys in `adata.layers`.")

            new_data.X = new_data.layers[layer]

        # DataFrame
        ul.log(__name__).info(f"to DataFrame")
        df: DataFrame = adata_map_df(new_data, column="value")

    elif isinstance(new_data, DataFrame):

        if legend_list is not None:
            df: DataFrame = new_data[new_data[label].isin(legend_list)].copy()
        else:
            df: DataFrame = new_data.copy()

    else:
        ul.log(__name__).error(f"The `data` parameter only support `AnnData` and `DataFrame` class types.")
        raise ValueError(f"The `data` parameter only support `AnnData` and `DataFrame` class types.")

    if legend is None:
        legend = "category"

    df[legend] = df[label].copy()
    new_data_columns = list(df.columns)

    hue_types = list(set(df[legend]))

    # noinspection DuplicatedCode
    if colors is not None:
        palette = colors
    else:
        if "color" in new_data_columns:
            palette = df["color"]
        else:
            palette = []

            for i in range(len(hue_types)):
                palette.append(plot_color_types[color_type][start_color_index + i * color_step_size + i])

    # sns.set_theme(style="whitegrid")
    ax.set(ylim=y_limit)
    sns.despine()

    if is_str:
        df[x] = df[x].astype(str)

    chart = sns.lineplot(data=df, ax=ax, x=x, y=y, hue=legend, palette=palette, linewidth=line_width, **kwargs)

    if is_str:
        locator = mdates.DayLocator(interval=1)
        chart.xaxis.set_major_locator(locator)

        ax.tick_params(axis='x', rotation=x_name_rotation)
    else:
        plt.xticks(x_ticks, rotation=x_name_rotation)

    plot_end(fig, title, x_name, y_name, output, show, close)



# -*- coding: UTF-8 -*-

import os
import pickle
import warnings
from pathlib import Path
from typing import Tuple, Optional

import h5py
import numpy as np
import pandas as pd
import anndata as ad
from tqdm import tqdm
from anndata import AnnData
from pandas import DataFrame

from .. import util as ul
from ..util import path, to_sparse, list_duplicate_set, collection, project_name, project_version

__name__: str = "file_read"


def read_h5ad(file: path, is_verbose: bool = True) -> AnnData:
    """
    Read AnnData data.
    :param file: file path;
    :param is_verbose: Set true to print log;
    :return: AnnData data.
    """
    if is_verbose:
        ul.log(__name__).info("Reading AnnData file: {}".format(file))

    return ad.read_h5ad(Path(file))


def read_h5(file: path, is_close: bool = False):
    """
    Read AnnData data
    :param file: file path
    :param is_close: file path
    :return: AnnData data
    """
    file = h5py.File(file, 'r')
    keys = file.keys()

    if is_close:
        file.close()

    return keys, file


def read_pkl(file: path, is_verbose: bool = True):
    """
    Read pkl data
    :param file: file path
    :param is_verbose: Set true to print log;
    :return: Python variable data
    """
    if is_verbose:
        ul.log(__name__).info("Reading pkl file: {}".format(file))

    # Recovering variables from files
    with open(str(file), 'rb') as f:
        data = pickle.load(f)

    return data


def handle_file_data_cell(
    file: path,
    clusters: str = "clusters",
    barcode_split_character: str = '-',
    on_barcode_split_character: str = None,
    is_transpose: bool = True,
    cluster_anno_file: path = None
) -> Tuple[DataFrame, DataFrame]:
    """
    Read table file to generate AnnData format data.
    :param file: A table with cell or peak column and column indexes, and the content is the number of fragments;
    :param clusters: The column name for cell clusters or cell types. (In most cases, this column can be ignored.)
        It is worth noting that only the values in this column are judged to determine whether they contain NA values;
        If they do, they are assigned the value `unknown`, and if not, no operation is performed;
    :param barcode_split_character: A barcode separated character symbol. (meta)
    :param on_barcode_split_character: A barcode separated character symbol. (matrix)
    :param is_transpose: Whether transpose is required to read the matrix file, default to True;
    :param cluster_anno_file: The file that adds information about cells must contain the column name `barcodes`;
    :return: Cell annotation data, counts data.
    """

    if file is None:
        ul.log(__name__).error("File cannot be empty")
        raise ValueError("File cannot be empty")

    if on_barcode_split_character is None:
        on_barcode_split_character = barcode_split_character

    data = pd.read_table(file, header=0)
    data = data.T if is_transpose else data
    # format barcodes information
    ul.log(__name__).info("handle cells information")
    cell_annot = pd.DataFrame(data=np.matrix(data.index).T, columns=["barcodes"])
    content_split_size = len(str(list(data.index)[0]).split(barcode_split_character))

    if content_split_size == 1:
        ul.log(__name__).info(f"Barcode does not contain a separator `{barcode_split_character}`")
        cell_annot["barcode"] = cell_annot["barcodes"]
    else:
        cell_annot: DataFrame = cell_annot["barcodes"].astype('str').str.split(barcode_split_character, expand=True)
        cell_annot.columns = ["barcode", "batch_id"]
        cell_annot["barcodes"] = (
            cell_annot["barcode"].astype(str)
            + on_barcode_split_character
            + cell_annot["batch_id"].astype(str)
        )

    # add annotation file
    if cluster_anno_file is not None:
        cell_annot = barcodes_add_anno(cluster_anno_file, cell_annot, clusters=clusters)

    cell_annot.index = cell_annot["barcode"].astype(str)
    return cell_annot, data


def barcodes_add_anno(annotation_file: path, cell_anno: DataFrame, clusters: str = None) -> DataFrame:
    """
    Add user inputted cell information
    :param annotation_file: The file that adds information about cells must contain the column name `barcodes`, the file input by the user.
    :param cell_anno: Read the cell description in the scATAC-seq data generated from the file.
    :param clusters: The column name for cell clusters or cell types. (In most cases, this column can be ignored.)
        It is worth noting that only the values in this column are judged to determine whether they contain NA values.
        If they do, they are assigned the value `unknown`, and if not, no operation is performed.
    :return: Complete cell annotation data
    """
    ul.log(__name__).info("Add annotation file")
    # add annotation file
    cell_annotation_file: DataFrame = pd.read_table(annotation_file, sep="\t", header=0, index_col=None)

    # judge column
    if "barcodes" not in list(cell_annotation_file.columns):
        ul.log(__name__).error("The annotation file must contain a column name with `barcodes`")
        raise ValueError("The annotation file must contain a column name with `barcodes`")

    if "barcodes" not in list(cell_anno.columns):
        ul.log(__name__).error("The `cell_anno` must contain a column name with `barcodes`")
        raise SyntaxError("The `cell_anno` must contain a column name with `barcodes`")

    if len(cell_anno["barcodes"]) > len(set(cell_annotation_file["barcodes"])):
        ul.log(__name__).error("Insufficient number of barcodes in the annotation file.")
        raise SyntaxError("Insufficient number of barcodes in the annotation file.")

    cell_anno = cell_anno.merge(cell_annotation_file, on="barcodes", how="inner")

    if clusters is not None:
        if clusters not in list(cell_annotation_file.columns):
            ul.log(__name__).error(f"The comment file must contain a column name with `{clusters}`, Try changing the `clusters` parameter")
            raise ValueError(f"The comment file must contain a column name with `{clusters}`, Try changing the `clusters` parameter")

        if clusters not in list(cell_anno.columns):
            cell_anno = cell_anno.rename(columns={f"{clusters}_y": clusters})

        # nan set unknown
        if cell_anno[cell_anno[clusters].isna()].shape[0] > 0:
            ul.log(__name__).warning(f"Due to the presence of `NA` in the `{clusters}`, it is forcibly assigned as `unknown`.")
            cell_anno.loc[cell_anno[clusters].isna(), clusters] = "unknown"

    if "barcode" not in list(cell_anno.columns):
        cell_anno = cell_anno.rename(columns={"barcode_x": "barcode"})

    cell_anno.index = cell_anno["barcode"].astype(str)
    return cell_anno


def read_barcodes_file(
    barcodes_file: path,
    clusters: str = None,
    barcode_split_character: str = '-',
    annotation_file: path = None,
) -> DataFrame:
    ul.log(__name__).info("handle cells information")
    # read file
    cell_anno = pd.read_csv(barcodes_file, header=None, index_col=None)
    content_split_size = len(str(cell_anno[0][1]).split(barcode_split_character))

    if content_split_size == 1:
        ul.log(__name__).info(f"Barcode does not contain a separator `{barcode_split_character}`")

    # judge
    if cell_anno.shape[0] != np.unique(cell_anno[0]).size:
        ul.log(__name__).error("Barcodes cannot have duplicate barcodes in the barcodes file")
        raise ValueError("Barcodes cannot have duplicate barcodes in the barcodes file")

    if barcode_split_character is None or content_split_size == 1:
        cell_anno.rename({0: "barcode"}, axis="columns", inplace=True)
        cell_anno["barcodes"] = cell_anno["barcode"].astype(str)
    else:
        __cell_annot__ = pd.read_csv(barcodes_file, sep=barcode_split_character, header=None, index_col=None)

        __is_with__ = True

        # judge
        if len(__cell_annot__.columns) == 1:
            ul.log(__name__).warning("Parameter `barcode_split_character` is `-`, but it is not working, ignore.")
            __is_with__ = False

        if __cell_annot__.shape[0] != np.unique(__cell_annot__[0]).size:
            ul.log(__name__).info("After extracting `batch_id`, there are duplicate barcodes, so `batch_id` is not extracted.")
            __is_with__ = False

        if __is_with__:
            cell_anno = __cell_annot__
            cell_anno.rename({0: "barcode", 1: "batch_id"}, axis="columns", inplace=True)
            cell_anno["barcodes"] = cell_anno["barcode"].astype(str) + "-" + cell_anno["batch_id"].astype(str)
        else:
            cell_anno.rename({0: "barcode"}, axis="columns", inplace=True)
            cell_anno["barcodes"] = cell_anno["barcode"].astype(str)

    # add annotation file
    if annotation_file is not None:
        cell_anno = barcodes_add_anno(annotation_file, cell_anno, clusters=clusters)

    cell_anno.set_index("barcode", inplace=True, drop=False)
    cell_anno.index = cell_anno.index.astype(str)
    return cell_anno


def _read_info_by_metadata_(
    base_path: path,
    feature_file_name: str,
    is_transpose: bool = True,
    clusters: str = None,
    barcode_split_character: str = '-',
    annotation_file: path = None,
) -> AnnData:
    """
    Read metadata outputted by 10x Genomics software.
    :param base_path: Path to directory with matrix, bed file, etc. (It can be obtained through cell-ranger)
    :param feature_file_name: feature file name;
    :param is_transpose: Whether transpose is required to read the matrix file, default to True;
    :param clusters: The column name for cell clusters or cell types. (In most cases, this column can be ignored.)
        It is worth noting that only the values in this column are judged to determine whether they contain NA values;
        If they do, they are assigned the value `unknown`, and if not, no operation is performed;
    :param barcode_split_character: A barcode separated character symbol. (meta)
    :param annotation_file: The file that adds information about cells must contain the column name `barcodes`;
    :return: sequence data
    """
    # read features file
    ul.log(__name__).info("handle features information")
    coords = pd.read_table(os.path.join(base_path, feature_file_name), header=None, index_col=None)
    coords.rename({0: "chr", 1: "start", 2: "end"}, axis="columns", inplace=True)
    coords.set_index(
        coords.chr.astype(str) + ":" + coords.start.astype(str) + "-" + coords.end.astype(str),
        inplace=True
    )
    coords.index = coords.index.astype(str)

    # read barcodes file
    cell_annot = read_barcodes_file(
        os.path.join(base_path, "barcodes.tsv"),
        clusters=clusters,
        barcode_split_character=barcode_split_character,
        annotation_file=annotation_file
    )

    from scipy.io import mmread

    # read matrix file
    ul.log(__name__).info("read `matrix.mtx` file")
    data = mmread(os.path.join(base_path, "matrix.mtx")).transpose() if is_transpose else \
        mmread(os.path.join(base_path, "matrix.mtx"))

    return AnnData(data.tocsr(), var=coords, obs=cell_annot)


def _process_peaks_(
    data: collection,
    peak_split_character: Tuple = (":", "-")
) -> Tuple[DataFrame, list, list, list]:
    """
    Processing peak set.
    :param data: Input peak set;
    :param peak_split_character: The connection symbol between chromosome, start and end;
    :return: Peak information, chromosome list, start position list, end position list.
    """
    # format peaks information
    ul.log(__name__).info("handle peaks information")
    features = pd.DataFrame(columns=["chr", "start", "end"])
    chr_list: list = []
    start_list: list = []
    end_list: list = []

    if peak_split_character[0] == peak_split_character[1]:
        character_ = peak_split_character[0]

        for col in list(data):
            col: str
            split: list = col.split(character_)
            chr_list.append(split[0])
            start_list.append(int(split[1]))
            end_list.append(int(split[2]))
    else:

        for col in list(data):
            col: str
            split: list = col.split(peak_split_character[0])
            chr_list.append(split[0])
            split2: list = split[1].split(peak_split_character[1])
            start_list.append(int(split2[0]))
            end_list.append(int(split2[1]))

    features["chr"] = chr_list
    features["start"] = start_list
    features["end"] = end_list
    features.set_index(
        features.chr.astype(str) + ":" + features.start.astype(str) + "-" + features.end.astype(str),
        inplace=True
    )
    features.index = features.index.astype(str)

    return features, chr_list, start_list, end_list


def collect_datasets(dsets: dict, group: h5py.Group):
    for k, v in group.items():
        if isinstance(v, h5py.Dataset):
            dsets[k] = v[()]
        else:
            collect_datasets(dsets, v)


def read_v3_10x_h5(filename: path) -> AnnData:
    """
    Read hdf5 file from Cell Ranger v3 or later versions.
    :param filename: H5 file
    :return: scATAC-seq data
    """
    ul.log(__name__).info('Start read hdf5 file')
    with h5py.File(str(filename), 'r') as f:
        try:
            dsets = {}
            collect_datasets(dsets, f["matrix"])

            from scipy.sparse import csr_matrix

            m, n = dsets['shape']
            data = dsets['data']
            if dsets['data'].dtype == np.dtype('int32'):
                data = dsets['data'].view('float32')
                data[:] = dsets['data']
            matrix = csr_matrix(
                (data, dsets['indices'], dsets['indptr']),
                shape=(n, m),
            )
            adata = AnnData(
                matrix,
                obs=dict(obs_names=dsets['barcodes'].astype(str)),
                var=dict(
                    var_names=dsets['name'].astype(str),
                    peak_ids=dsets['id'].astype(str),
                    feature_types=dsets['feature_type'].astype(str),
                    genome=dsets['genome'].astype(str),
                ),
            )
            return adata
        except KeyError:
            raise Exception('File is missing one or more required datasets.')


def read_sc_atac_10x_h5(
    file: path,
    clusters: str = None,
    barcode_split_character: str = '-',
    annotation_file: path = None,
    peak_split_character: Tuple = (":", "-")
) -> AnnData:
    """
    Read hdf5 file from Cell Ranger v3 or later versions.
    :param file: A comprehensive h5ad file. (It can be obtained through cell-ranger)
    :param clusters: The column name for cell clusters or cell types. (In most cases, this column can be ignored.)
        It is worth noting that only the values in this column are judged to determine whether they contain NA values.
        If they do, they are assigned the value `unknown`, and if not, no operation is performed.
    :param barcode_split_character: A barcode separated character symbol (meta)
    :param annotation_file: The file that adds information about cells must contain the column name `barcodes`
    :param peak_split_character: A peak separated character symbol
    :returns: scATAC-seq data
    """
    ul.log(__name__).info("Reading scATAC-seq 10x data")
    sc_atac: AnnData = read_v3_10x_h5(file)
    # cell information
    cells = sc_atac.obs.copy()
    barcodes = list(cells.index)
    cells["barcodes"] = barcodes

    # Add bath
    if len(str(barcodes[0]).split(barcode_split_character)) == 2:
        barcode_list: list = []
        batch_id_list: list = []

        for barcode in barcodes:
            barcode: str
            barcode_list.append(barcode.split(barcode_split_character)[0])
            batch_id_list.append(barcode.split(barcode_split_character)[1])

        cells["barcode"] = barcode_list
        cells["batch_id"] = batch_id_list
    else:
        cells["barcode"] = barcodes

    # add annotation file
    if annotation_file is not None:
        cells = barcodes_add_anno(annotation_file, cells, clusters=clusters)

    sc_atac.obs = cells

    # peak information
    peaks: DataFrame = sc_atac.var.copy()
    # format peaks information
    _, chr_list, start_list, end_list = _process_peaks_(peaks.index, peak_split_character)
    peaks["chr"] = chr_list
    peaks["start"] = start_list
    peaks["end"] = end_list
    sc_atac.var = peaks

    return sc_atac


def read_sc_atac(
    resource: path = None,
    is_transpose: bool = True,
    barcode_split_character: str = '-',
    on_barcode_split_character: str = None,
    annotation_file: path = None,
    clusters: str = None,
    peak_split_character: Tuple = (":", "-")
) -> AnnData:
    """
    Return scATAC-seq data in AnnData format.
    :param resource:
        1. Path to directory with matrix, bed file, etc. (It can be obtained through cell-ranger)
        2. H5 file obtained through cell-ranger.
        3. A comprehensive h5ad file.
        4. A table with cell or peak column and column indexes, and the content is the number of fragments;
    :param is_transpose: Whether transpose is required to read the matrix file, default to True;
    :param barcode_split_character: A barcode separated character symbol. (meta)
    :param on_barcode_split_character: A barcode separated character symbol. (matrix)
    :param annotation_file: The file that adds information about cells must contain the column name `barcodes`;
    :param clusters: The column name for cell clusters or cell types. (In most cases, this column can be ignored.)
        It is worth noting that only the values in this column are judged to determine whether they contain NA values;
        If they do, they are assigned the value `unknown`, and if not, no operation is performed;
    :param peak_split_character: A peak separated character symbol;
    :return: scATAC-seq data.
    """
    ul.log(__name__).info("Read scATAC-seq data")

    is_metadata: bool = os.path.isdir(str(resource))

    if is_metadata:

        sc_atac: AnnData = _read_info_by_metadata_(
            base_path=resource,
            feature_file_name="peaks.bed",
            is_transpose=is_transpose,
            clusters=clusters,
            barcode_split_character=barcode_split_character,
            annotation_file=annotation_file
        )

    else:

        if str(resource).endswith(".h5"):
            sc_atac: AnnData = read_sc_atac_10x_h5(
                file=resource,
                barcode_split_character=barcode_split_character,
                peak_split_character=peak_split_character
            )

            # add annotation file
            if annotation_file is not None:
                cell_annot = sc_atac.obs.copy()
                cell_annot = barcodes_add_anno(annotation_file, cell_annot, clusters=clusters)
                sc_atac.obs = cell_annot

        elif str(resource).endswith(".h5ad"):
            sc_atac: AnnData = read_h5ad(file=resource)

            peak_col: list = list(sc_atac.var.columns)
            barcode_col: list = list(sc_atac.obs.columns)

            if "chr" not in peak_col or "start" not in peak_col or "end" not in peak_col:
                _, chr_list, start_list, end_list = _process_peaks_(list(sc_atac.var_names), peak_split_character)

                if len(peak_split_character) != 2:
                    ul.log(__name__).error(
                        "The peak feature is used to obtain the segmentation character of `chr` `start` `end`, "
                        "which requires two characters. The first character is used to segment `chr` and `start`, "
                        "and the second character is used to segment `start` and `end`"
                    )
                    raise ValueError(
                        "The peak feature is used to obtain the segmentation character of `chr` `start` `end`, "
                        "which requires two characters. The first character is used to segment `chr` and `start`, "
                        "and the second character is used to segment `start` and `end`"
                    )

                sc_atac.var["chr"] = chr_list
                sc_atac.var["start"] = start_list
                sc_atac.var["end"] = end_list

            if "barcode" not in barcode_col and "barcodes" in barcode_col:
                sc_atac.obs["barcode"] = sc_atac.obs["barcodes"]
            elif "barcode" in barcode_col and "barcodes" not in barcode_col:
                sc_atac.obs["barcodes"] = sc_atac.obs["barcode"]
            elif "barcode" not in barcode_col and "barcodes" not in barcode_col:
                sc_atac.obs["barcode"] = sc_atac.obs_names
                sc_atac.obs["barcodes"] = sc_atac.obs_names

            # add annotation file
            if annotation_file is not None:
                cell_annot = sc_atac.obs.copy()
                cell_annot = barcodes_add_anno(annotation_file, cell_annot, clusters=clusters)
                sc_atac.obs = cell_annot

        else:
            if len(peak_split_character) != 2:
                ul.log(__name__).error(
                    "The peak feature is used to obtain the segmentation character of `chr` `start` `end`, "
                    "which requires two characters. The first character is used to segment `chr` and `start`, "
                    "and the second character is used to segment `start` and `end`"
                )
                raise ValueError(
                    "The peak feature is used to obtain the segmentation character of `chr` `start` `end`, "
                    "which requires two characters. The first character is used to segment `chr` and `start`, "
                    "and the second character is used to segment `start` and `end`"
                )

            cell_annot, data = handle_file_data_cell(
                file=resource,
                clusters=clusters,
                barcode_split_character=barcode_split_character,
                on_barcode_split_character=on_barcode_split_character,
                is_transpose=is_transpose,
                cluster_anno_file=annotation_file
            )

            # format peaks information
            features, _, _, _ = _process_peaks_(data.columns, peak_split_character)
            sc_atac = AnnData(to_sparse(data.values), var=features, obs=cell_annot)

    # Add project
    sc_atac.uns["project_name"] = project_name
    sc_atac.uns["project_version"] = project_version
    # save params information
    sc_atac.uns["params"] = {
        "resource": resource,
        "is_transpose": is_transpose,
        "clusters": clusters,
        "barcode_split_character": barcode_split_character,
        "on_barcode_split_character": on_barcode_split_character,
        "annotation_file": annotation_file,
        "peak_split_character": {
            "chr_start_split_symbal": peak_split_character[0],
            "start_end_split_symbal": peak_split_character[1]
        },
        "is_metadata": is_metadata
    }

    return sc_atac


def read_variants(
    base_path: Optional[path] = None,
    files: Optional[collection] = None,
    labels: Optional[dict] = None,
    column_map: Optional[dict] = None,
    repeat_symbol: str = "_#"
) -> Tuple[dict, DataFrame]:
    """
    Read variant file set
    :param base_path: Path for storing mutation trait data;
        The file must contain the following names `{chr, position, rsId, pp}`, where ID represents the representative of the trait name;
    :param files: Collection of mutation trait data;
    :param labels: Classification labels for each trait or disease;
    :param column_map: The mapping of column names facilitates mapping the corresponding column names in the mutation file to the specified column name information,
        i.e. `{0: "chr", 1: "position", 2: "rsId", 3: "pp"}`;
    :param repeat_symbol: If you encounter two names with the same abbreviation, you can add a symbol and numerical value to one of the abbreviations.
    :return: Data on traits or diseases
        1. PP data for each trait or disease。
        2. Annotated information on traits or diseases。
    """
    ul.log(__name__).info("Read variants file information")
    variant_columns: list = ["chr", "position", "rsId", "pp"]
    need_variant_columns: list = variant_columns.copy()
    variant_columns.append("id")

    results: dict = {}
    filenames: dict = {}

    # get files
    file_size: int = 0

    if files is not None:
        file_size: int = len(list(files))

    if base_path is None and file_size == 0:
        ul.log(__name__).error("At least one of the `resource` and `files` parameters has a parameter")
        raise ValueError("At least one of the `resource` and `files` parameters has a parameter")

    new_files = []

    if isinstance(base_path, path):
        new_files = ul.file_method(__name__).get_files_path(base_path)

    if file_size > 0:
        new_files.extend(files)

    key_list: list = []

    # read file
    for variant_file in tqdm(new_files):

        key, _ = os.path.splitext(os.path.basename(variant_file))

        if column_map is not None:

            keys = list(column_map.keys())

            if isinstance(keys[0], str):
                variant = pd.read_table(variant_file, header=0, index_col=None)
            elif isinstance(keys[0], int):
                variant = pd.read_table(variant_file, header=None, index_col=None)
            else:
                raise ValueError(
                    "The `keys` in `column_map` only support `None`, `str` type, and `int` type. "
                    "When it is None or `str` type, it will be read with a header, and when it is `int` type, "
                    "it will be read without a header."
                )

            variant.rename(columns=column_map, inplace=True)
        else:
            variant = pd.read_table(variant_file, header=0, index_col=None)

        columns = variant.columns

        if not set(need_variant_columns).issubset(set(columns)):
            ul.log(__name__).error(f"The column name of the {variant_file} file needs to include {need_variant_columns}")
            raise ValueError(f"The column name of the {variant_file} file needs to include {need_variant_columns}")

        if variant.shape[0] == 0:
            ul.log(__name__).warning(f"The content of file {variant_file} is empty")
            continue

        # judge
        if key not in key_list:
            key_list.append(key)
            variant["id"] = key
        else:
            _label_: str = key.split(repeat_symbol)[-1]
            _number_: int = 2

            if _label_.isnumeric():
                _number_: int = int(_label_)
                _number_ += 1

            old_key: str = key
            key += f"{repeat_symbol}{str(_number_)}"
            ul.log(__name__).warning(f"The `{old_key}` already exists in the previous file, set this key to `{key}`")
            variant["id"] = key

        results.update({key: variant[variant_columns]})
        filenames.update({key: variant_file})

    ul.log(__name__).info("Read variants file finish")
    ul.log(__name__).info("handle variants file")
    # concat files
    results_info = pd.concat(list(results.values()), axis=0)
    results_info.reset_index()

    # format trait information
    pp_sum: DataFrame = results_info.groupby("id", as_index=False)["pp"].sum()
    pp_sum.columns = ["id", "pp_sum"]
    pp_mean = results_info.groupby("id", as_index=False)["pp"].mean()
    pp_mean.columns = ["id", "pp_mean"]
    pp_size = results_info.groupby("id", as_index=False)["rsId"].size()
    pp_size.columns = ["id", "count"]

    merge_info = pp_sum.merge(pp_mean, on="id", how="inner")
    trait_info: DataFrame = merge_info.merge(pp_size, on="id", how="inner")
    trait_info.index = trait_info["id"].astype(str)

    # Add filename
    trait_info["filename"] = filenames

    # format variant data
    variant_data_dict: dict = {}

    ul.log(__name__).info("Create multiple trait-variants data")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        for key in tqdm(trait_info.index):
            variant_content: DataFrame = results[key]
            # format variant information
            variant_content["variant_id"] = (
                variant_content["chr"].astype(str)
                + ":" + variant_content["position"].astype(str)
                + ":" + variant_content["rsId"].astype(str)
            )
            variant_info = variant_content[["variant_id", "chr", "position", "rsId"]]
            variant_info_index = list(variant_info["variant_id"].astype(str))
            # set index
            variant_content["variant_id_index"] = list_duplicate_set(variant_info_index)
            variant_info.index = variant_content["variant_id_index"]
            # format trait-variant data
            variant_list: list = list(variant_content["variant_id_index"])
            trait_name = list(variant_content["id"])[0]
            trait: DataFrame = trait_info[trait_info["id"] == trait_name]

            # format dict
            variant_dict: dict = dict(zip(variant_list, range(len(variant_list))))

            shape = (len(variant_list), 1)

            matrix = np.zeros(shape)

            for variant_id_index, pp in zip(variant_content["variant_id_index"], variant_content["pp"]):
                matrix[variant_dict[variant_id_index], 0] = pp

            # format AnnData
            variant_adata = AnnData(to_sparse(matrix), var=trait, obs=variant_info)
            variant_data_dict.update({trait_name: variant_adata})

    if labels is not None:
        trait_info["labels"] = labels
    else:
        trait_info["labels"] = trait_info["id"]

    return variant_data_dict, trait_info



# -*- coding: UTF-8 -*-

from typing import Optional, Literal

import pandas as pd
from pandas import DataFrame
from anndata import AnnData
import numpy as np

from .. import util as ul
from ..util import collection, matrix_data, to_sparse, to_dense, number, check_adata_get

__name__: str = "preprocessing_anndata"


def adata_group(
    adata: AnnData,
    column: str,
    extra_column: Optional[str] = None,
    axis: Literal[0, 1] = 1,
    layer: str = None,
    method: collection | str = ("mean", "sum", "median")
) -> AnnData:
    """
    Return reshaped AnnData organized by given `column` values.
    :param adata: input data;
    :param column: grouping `column`;
    :param extra_column: Extra columns reserved based on grouped `column`;
    :param axis: Which dimension is used for grouping. {1: adata.obs, 0: adata.var};
    :param layer: Specify the matrix to be processed;
    :param method: The method of grouping strategy supports the following 5 types and their combinations.
        The five methods are {"mean", "sum", "median", "max", "min"}.
    :return: Data grouped by AnnData.
    """
    # judge input data
    if adata.shape[0] == 0:
        ul.log(__name__).warning("Input data is empty")
        return adata

    # judge axis
    if not isinstance(axis, number) or axis not in range(2):
        ul.log(__name__).error("The `axis` parameter must be either 0 or 1")
        raise ValueError("The `axis` parameter must be either 0 or 1")

    # get data
    data: AnnData = adata.copy() if axis == 1 else adata.copy().T

    # judge layers
    if layer is not None:
        if layer not in list(data.layers):
            ul.log(__name__).error("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
            raise ValueError("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
        data.X = data.layers[layer]

    # get group information
    data_obs: DataFrame = data.obs
    if column not in data_obs.columns:
        ul.log(__name__).error(f"The grouped column {column} are not in the corresponding columns {data_obs.columns}")
        raise ValueError(f"The grouped column {column} are not in the corresponding columns {data_obs.columns}")

    if extra_column is not None and extra_column not in data_obs.columns:
        ul.log(__name__).error(f"The grouped extra-column {extra_column} are not in the corresponding columns {data_obs.columns}")
        raise ValueError(
            f"The grouped extra-column {extra_column} are not in the corresponding columns {data_obs.columns}"
        )

    # handle group information
    if extra_column is not None:
        obs: DataFrame = pd.DataFrame(data_obs[[extra_column, column]]).drop_duplicates()
        obs.index = obs[column].astype(str)
    else:
        obs: DataFrame = pd.DataFrame(data_obs[[column]]).drop_duplicates()
        obs.index = obs[column].astype(str)

    obs.sort_index(inplace=True)
    obs.rename_axis("group_column_index", inplace=True)
    column_size = obs.shape[0]
    column_group = list(obs[column])
    # create container
    matrix_mean: matrix_data = np.zeros((column_size, data.shape[1]))
    matrix_sum: matrix_data = np.zeros((column_size, data.shape[1]))
    matrix_max: matrix_data = np.zeros((column_size, data.shape[1]))
    matrix_min: matrix_data = np.zeros((column_size, data.shape[1]))
    matrix_median: matrix_data = np.zeros((column_size, data.shape[1]))

    # add data
    for i in range(column_size):
        # 获取 data_obs 下的索引信息
        data_obs_column: DataFrame = data_obs[data_obs[column] == column_group[i]]
        # sum value
        overlap_variant = data[list(data_obs_column.index), :]

        if "mean" in method:
            matrix_mean[i] = overlap_variant.X.mean(axis=0)

        if "sum" in method:
            matrix_sum[i] = overlap_variant.X.sum(axis=0)

        if "max" in method:
            matrix_max[i] = np.amax(to_dense(overlap_variant.X, is_array=True), axis=0)

        if "min" in method:
            matrix_min[i] = np.amin(to_dense(overlap_variant.X, is_array=True), axis=0)

        if "median" in method:
            matrix_median[i] = np.median(to_dense(overlap_variant.X, is_array=True), axis=0)

    # create result
    ann_data = AnnData(to_sparse(matrix_mean), obs=obs, var=data.var)

    if "sum" in method:
        ann_data.layers["sum"] = to_sparse(matrix_sum)

    if "max" in method:
        ann_data.layers["max"] = to_sparse(matrix_max)

    if "min" in method:
        ann_data.layers["min"] = to_sparse(matrix_min)

    if "median" in method:
        ann_data.layers["median"] = to_sparse(matrix_median)

    return ann_data if axis == 1 else ann_data.T


def adata_map_df(
    adata: AnnData,
    column: str = "value",
    layer: str = None
) -> DataFrame:
    """
    Convert AnnData to a form of `row   column  value`
    :param adata: Enter the AnnData data to be converted;
    :param column: Specify the column name of the value;
    :param layer: Specify the matrix to be processed;
    :return: The DataFrame data of the `row   column  value`.
    """
    # judge input data
    data: AnnData = check_adata_get(adata, layer=layer)

    # get group information
    data_obs: DataFrame = data.obs.copy()
    data_var: DataFrame = data.var.copy()

    if column in data_obs.columns or column in data_var.columns:
        ul.log(__name__).error(f"The newly generated column cannot be within the existing column name")
        raise ValueError(f"The newly generated column cannot be within the existing column name")

    # rename index
    __on__: str = "on_5645465353221"
    data_var.rename_axis("y_index", inplace=True)
    data_var.reset_index(inplace=True)
    data_var["on_"] = __on__
    data_obs.rename_axis("x_index", inplace=True)
    data_obs.reset_index(inplace=True)
    data_obs["on_"] = __on__

    # create data
    ul.log(__name__).info("Create Table")
    data_df: DataFrame = data_var.merge(data_obs, on="on_", how="outer")
    data_df.drop(["on_"], axis=1, inplace=True)
    data_df[column] = to_dense(data.X.T, is_array=True).flatten()
    return data_df



# -*- coding: UTF-8 -*-

import os
import pickle
import shutil
from typing import Union, Literal

import numpy as np
from scipy import sparse
from tqdm import tqdm

from ..file import save_pkl
from .. import util as ul
from ..util import (
    to_dense,
    to_sparse,
    sparse_matrix,
    matrix_data,
    generate_str,
    collection,
    file_method,
    number,
    dense_data
)

__name__: str = "tool_matrix"


def split_matrix(data: matrix_data, axis: Literal[0, 1] = 0, chunk_number: int = 1000) -> list:
    # get size
    new_data = to_dense(data, is_array=True)
    rows, cols = new_data.shape

    # get number
    total: int = rows if axis == 0 else cols

    # get split number
    split_number = total // chunk_number

    # Determine whether to divide equally
    tail_number = total % chunk_number
    if tail_number != 0:
        split_number += 1

    ul.log(__name__).info(f"Divide the matrix into {split_number} parts")

    # Add data
    split_data_list = []
    for i in range(split_number):
        # set index
        start_index: int = i * chunk_number
        end_index: int = total if (i + 1) * chunk_number > total else (i + 1) * chunk_number
        split_data = new_data[start_index:end_index, :] if axis == 0 else new_data[:, start_index:end_index]
        split_data_list.append(split_data)

    return split_data_list


def merge_matrix(datas: list, axis: Literal[0, 1] = 0) -> list:
    # get size
    size = len(datas)
    range_size = range(size)

    # get row col
    constant: int = datas[0].shape[1] if axis == 0 else datas[0].shape[0]
    total: int = 0
    shapes: list = []

    # get chunk size
    for i in range_size:
        shape = datas[i].shape
        judge = shape[1] if axis == 0 else shape[0]

        # judge size
        if judge != constant:
            ul.log(__name__).error("Inconsistent traits in the input dataset set.")
            raise ValueError("Inconsistent traits in the input dataset set.")

        total += shape[0] if axis == 0 else shape[1]
        shapes.append(shape)

    # format matrix
    matrix_shape = (total if axis == 0 else constant, constant if axis == 0 else total)
    ul.log(__name__).info(f"Merge matrix {matrix_shape} shape")
    matrix: matrix_data = np.zeros(matrix_shape)

    # merge matrix
    col_record: int = 0
    for i in range_size:
        col_i = datas[i].shape[0] if axis == 0 else datas[i].shape[1]

        if axis == 0:
            matrix[col_record:col_record + col_i, :] = datas[i]
        else:
            matrix[:, col_record:col_record + col_i] = datas[i]

        col_record += col_i

    return matrix


def down_sampling_data(data: Union[matrix_data | collection], sample_number: int = 1000000) -> list:
    """
    down-sampling
    :param data: Data that requires down-sampling;
    :param sample_number: How many samples (values) were down-sampled.
    :return: Data after down-sampling.
    """

    if isinstance(data, collection):

        # Judge data size
        if data.size <= sample_number:
            return list(data)

        index = np.random.choice(range(data.size), sample_number, replace=False)

        return list(np.array(data)[index])

    elif isinstance(data, matrix_data):

        # Judge data size
        if data.shape[0] * data.shape[1] <= sample_number:
            return list(to_dense(data, is_array=True).flatten())

        data = to_dense(data, is_array=True)
        row_count = data.shape[0]
        col_count = data.shape[1]

        if row_count < 0:
            ul.log(__name__).error("The number of rows of data must be greater than zero")
            raise ValueError("The number of rows of data must be greater than zero")

        ul.log(__name__).info(
            f"Kernel density estimation plot down-sampling data from {row_count * col_count} to {sample_number}"
        )

        # get count
        count = row_count * col_count
        iter_number: int = count // sample_number
        iter_sample_number: int = sample_number // iter_number
        iter_sample_number_final: int = sample_number % iter_number

        if iter_sample_number < 1:
            ul.log(__name__).error("The sampling data is too small, increase the `sample_number` parameter value")
            raise ValueError("The sampling data is too small, increase the `sample_number` parameter value")

        ul.log(__name__).info(f"Divide and conquer {iter_number} chunks")

        # Create index container
        return_data: list = []

        for i in range(iter_number + 1):

            if iter_number < 50:
                ul.log(__name__).info(f"Start {i + 1}th chunk, {(i + 1) / iter_number * 100}%")
            elif iter_number >= 50 and i % 50 == 0:
                ul.log(__name__).info(f"Start {i + 1}th chunk, {(i + 1) / iter_number * 100}%")

            # Determine if it is the last cycle
            end_count: int = count if i == iter_number else (i + 1) * sample_number

            if iter_sample_number_final == 0:
                index = np.random.choice(range(i * sample_number, end_count), iter_sample_number, replace=False)
            else:
                per_iter_sample_number: int = iter_sample_number_final if i == iter_number else iter_sample_number
                index = np.random.choice(range(i * sample_number, end_count), per_iter_sample_number, replace=False)

            # Add index
            for j in index:
                # row
                row_index = j // col_count
                # column
                col_index = j % col_count

                if row_index >= row_count:
                    ul.log(__name__).error(f"index ({row_index}) out of range ({row_count})")
                    raise IndexError(f"index ({row_index}) out of range ({row_count})")

                if col_index >= col_count:
                    ul.log(__name__).error(f"index ({col_index}) out of range ({col_count})")
                    raise IndexError(f"index ({col_index}) out of range ({col_count})")

                return_data.append(data[row_index, col_index])

        return return_data
    else:
        ul.log(__name__).error("The input data type is incorrect.")
        raise ValueError("The input data type is incorrect.")


def matrix_dot_block_storage(
    data1: matrix_data,
    data2: matrix_data,
    block_size: int = 10000,
    is_return_sparse: bool = False,
    data: matrix_data = None
) -> matrix_data:
    """
    Perform Cartesian product of two matrices through block storage method.
    :param data1: Matrix 1
    :param data2: Matrix 2
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param is_return_sparse: Whether to return sparse matrix
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Cartesian product result
    """

    n, m = data1.shape
    p, q = data2.shape

    if m != p:
        ul.log(__name__).error(f"Need to meet the principle of matrix multiplication. ({m} != {p})")
        raise ValueError(f"Need to meet the principle of matrix multiplication. ({m} != {p})")

    if block_size <= 0 or n <= block_size and m <= block_size and q <= block_size:
        return to_sparse(np.dot(data1, data2)) if is_return_sparse else np.dot(data1, data2)

    n_range = range(0, n, block_size)
    q_range = range(0, q, block_size)
    m_range = range(0, m, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    ul.log(__name__).info("[matrix_dot_block_storage]: Store block data...")
    total_steps = len(n_range) * len(q_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                for k in m_range:
                    i_max = min(i + block_size, n)
                    j_max = min(j + block_size, q)
                    k_max = min(k + block_size, m)

                    _matrix_ = np.dot(data1[i:i_max, k:k_max], data2[k:k_max, j:j_max]).astype(np.float32)
                    save_pkl(_matrix_, os.path.join(_cache_path_, str(i) + str(j) + str(k) + ".pkl"))
                    del _matrix_

                    pbar.update(1)

    del data1, data2

    if data is None or data.shape != (n, q):
        data = sparse.lil_matrix((n, q)).astype(np.float32) if is_return_sparse else np.zeros((n, q)).astype(np.float32)

    # Read data
    ul.log(__name__).info("[matrix_dot_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                for k in m_range:
                    i_max = min(i + block_size, n)
                    j_max = min(j + block_size, q)

                    with open(os.path.join(_cache_path_, str(i) + str(j) + str(k) + ".pkl"), 'rb') as f:
                        _matrix_ = pickle.load(f)
                        data[i:i_max, j:j_max] += _matrix_
                        del _matrix_

                    pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data.tocsr() if is_return_sparse else data


def matrix_multiply_block_storage(
    data1: matrix_data,
    data2: matrix_data,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Perform Hadamard product of two matrices through block storage method.
    :param data1: Matrix 1
    :param data2: Matrix 2
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Hadamard product result
    """
    n, m = data1.shape
    n1, m1 = data2.shape

    if n != n1 or m != m1:
        ul.log(__name__).error(
            f"Need to satisfy the multiplication principle of Hadamard products in matrices. ({(n, m)} != {n1, m1})"
        )
        raise ValueError(
            f"Need to satisfy the multiplication principle of Hadamard products in matrices. ({(n, m)} != {n1, m1})"
        )

    if block_size <= 0 or n <= block_size and m <= block_size:
        return np.multiply(data1, data2)

    n_range = range(0, n, block_size)
    m_range = range(0, m, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    ul.log(__name__).info("[matrix_multiply_block_storage]: Store block data...")
    total_steps = len(n_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                _matrix_ = np.multiply(data1[i:i_max, k:k_max], data2[i:i_max, k:k_max])
                save_pkl(_matrix_, os.path.join(_cache_path_, str(i) + str(k) + ".pkl"))
                del _matrix_

                pbar.update(1)

    del data1, data2

    if data is None or data.shape != (n, m):
        data = np.zeros((n, m))

    # Read data
    ul.log(__name__).info("[matrix_multiply_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, k:k_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def matrix_operation_memory_efficient(
    data1: matrix_data,
    data2: Union[matrix_data, number],
    chunk_size: int = 10000,
    default: float = 1e+8,
    operation: Literal['+', '-', '*', '/'] = '*'
) -> sparse_matrix:
    """
    Perform element-wise addition, subtraction, multiplication, and division on two sparse matrices by blocks, supporting memory-efficient processing.
    :param data1: Sparse matrix 1
    :param data2: Sparse matrix 2
    :param chunk_size: The size of the segmentation stored in block wise element-wise operation.
        If the value is less than or equal to zero, no block operation will be performed.
    :param default: 1e+8
    :param operation: Element-wise operation type, optional '+', '-', '*', '/'
    :return: Result sparse matrix (CSR format)
    """
    n, m = data1.shape

    if isinstance(data2, matrix_data):

        n1, m1 = data2.shape

        # Element-wise operations follow below
        if n != n1 or m != m1:
            ul.log(__name__).error(f"Need to satisfy the element-wise operation principle in matrices. ({(n, m)} != {n1, m1})")
            raise ValueError(f"Need to satisfy the element-wise operation principle in matrices. ({(n, m)} != {n1, m1})")
    else:

        if operation == "/" and data2 == 0:

            if default == 0:
                ul.log(__name__).error(f"The denominator (`data2`) cannot be zero.")
                raise ValueError(f"The denominator (`data2`) cannot be zero.")

            ul.log(__name__).warning(f"The denominator (`data2`) cannot be zero, it defaults to the `default` parameter value.")
            data2 = default

    if chunk_size <= 0:

        if operation == '+':

            if isinstance(data1, matrix_data):
                data2 = to_dense(data2) if isinstance(data1, dense_data) else to_sparse(data2)
                return data1 + data2
            else:
                data1.data = data1.data + data2
                return data1

        elif operation == '-':

            if isinstance(data1, matrix_data):
                data2 = to_dense(data2) if isinstance(data1, dense_data) else to_sparse(data2)
                return data1 - data2
            else:
                data1.data = data1.data - data2
                return data1

        elif operation == '*':

            if isinstance(data1, matrix_data):
                data2 = to_dense(data2) if isinstance(data1, dense_data) else to_sparse(data2)
                return data1.multiply(data2)
            else:
                data1.data = data1.data * data2
                return data1

        elif operation == '/':

            if isinstance(data1, matrix_data):
                data1 = to_dense(data1)
                data2 = to_dense(data2)
                data2[data2 == 0] = default
                return (data1 / data2) if isinstance(data1, dense_data) else to_sparse(data1 / data2)
            else:
                data1.data = data1.data / data2
                return data1

        else:
            ul.log(__name__).error(f"Unsupported operation: {operation}")
            raise ValueError(f"Unsupported operation: {operation}")

    if isinstance(data1, matrix_data):
        data2 = to_dense(data2) if isinstance(data1, dense_data) else to_sparse(data2)

    result = sparse.lil_matrix(data1.shape).astype(np.float32)

    for i in tqdm(range(0, data1.shape[0], chunk_size)):
        i_end = min(i + chunk_size, data1.shape[0])

        chunk1 = data1[i:i_end, :]

        if isinstance(data2, matrix_data):
            chunk2 = data2[i:i_end, :]
        else:
            chunk2 = data2

        if operation == '+':
            result_chunk = chunk1 + chunk2
        elif operation == '-':
            result_chunk = chunk1 - chunk2
        elif operation == '*':
            result_chunk = chunk1.multiply(chunk2)
        elif operation == '/':
            # Sparse matrix division: convert to dense, perform element-wise division, then convert back to sparse
            dense_chunk1 = chunk1 if isinstance(chunk1, dense_data) else chunk1.todense()

            if isinstance(data2, matrix_data):
                dense_chunk2 = chunk2 if isinstance(chunk2, dense_data) else chunk2.todense()
                dense_chunk2[dense_chunk2 == 0] = default
            else:
                dense_chunk2 = data2

            result_chunk = dense_chunk1 / dense_chunk2
            result_chunk = sparse.csr_matrix(result_chunk)
        else:
            ul.log(__name__).error(f"Unsupported operation: {operation}")
            raise ValueError(f"Unsupported operation: {operation}")

        result[i:i_end, :] = result_chunk

    return result.tocsr()


def vector_multiply_block_storage(
    data1: collection,
    data2: collection,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Two vectors are broadcast in rows and columns respectively and multiplied by Hadamard product
    :param data1: Vector 1
    :param data2: Vector 2
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Matrix
    """

    vector1 = to_dense(data1, is_array=True).flatten()[:, np.newaxis]
    vector2 = to_dense(data2, is_array=True).flatten()

    del data1, data2

    n, _ = vector1.shape
    q = vector2.shape[0]

    if block_size <= 0 or n <= block_size and q <= block_size:
        return vector1 * vector2

    n_range = range(0, n, block_size)
    q_range = range(0, q, block_size)

    # block data
    ul.log(__name__).info("[vector_multiply_block_storage]: Block calculation...")

    if data is None or data.shape != (n, q):
        data = np.zeros((n, q))

    total_steps = len(n_range) * len(q_range)

    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                i_max = min(i + block_size, n)
                j_max = min(j + block_size, q)

                data[i:i_max, j:j_max] += (vector1[i:i_max, :] * vector2[j:j_max]).astype(np.float32)

                pbar.update(1)

    del vector1, vector2

    return data


def matrix_division_block_storage(
    matrix: matrix_data,
    value: Union[float, int, collection, matrix_data],
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Dividing a matrix by another value, vector, or matrix.
    :param matrix: Matrix
    :param value: Value, vector, or matrix
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Matrix
    """

    n, m = matrix.shape

    is_status: int = -1

    if isinstance(value, collection):
        value = to_dense(value, is_array=True)
        q = value.shape[0]

        if q != n and q != m:
            ul.log(__name__).error(f"Inconsistent dimensions cannot be divided. {(n, m)} != {q}")
            raise ValueError(f"Inconsistent dimensions cannot be divided. {(n, m)} != {q}")

        is_status = 1 if n == q else 0
    elif isinstance(value, matrix_data):
        is_status = 2

    if block_size <= 0 or n <= block_size and m <= block_size:
        return matrix / value

    n_range = range(0, n, block_size)
    m_range = range(0, m, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    ul.log(__name__).info("[matrix_division_block_storage]: Store block data...")
    total_steps = len(n_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                if is_status == -1:
                    _matrix_ = matrix[i:i_max, k:k_max] / value
                elif is_status == 0:
                    _matrix_ = matrix[i:i_max, k:k_max] / value[k:k_max]
                elif is_status == 1:
                    _matrix_ = matrix[i:i_max, k:k_max] / value[i:i_max]
                elif is_status == 2:
                    _matrix_ = matrix[i:i_max, k:k_max] / value[i:i_max, k:k_max]
                else:
                    raise RuntimeError("nothingness")

                save_pkl(_matrix_, os.path.join(_cache_path_, str(i) + str(k) + ".pkl"))
                del _matrix_

                pbar.update(1)

    del matrix, value

    if data is None or data.shape != (n, m):
        data = np.zeros((n, m))

    # Read data
    ul.log(__name__).info("[matrix_division_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, k:k_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def matrix_callback_block_storage(
    matrix: matrix_data,
    callback,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Dividing a matrix by another value, vector, or matrix.
    :param matrix: Matrix
    :param callback: callback function
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Matrix
    """

    n, m = matrix.shape

    n_range = range(0, n, block_size)
    m_range = range(0, m, block_size)

    if block_size <= 0 or n <= block_size and m <= block_size:
        return callback(matrix)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    ul.log(__name__).info("[matrix_callback_block_storage]: Store block data...")
    total_steps = len(n_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                _matrix_ = callback(matrix[i:i_max, k:k_max])

                save_pkl(_matrix_, os.path.join(_cache_path_, str(i) + str(k) + ".pkl"))
                del _matrix_

                pbar.update(1)

    del matrix

    if data is None or data.shape != (n, m):
        data = np.zeros((n, m))

    # Read data
    ul.log(__name__).info("[matrix_callback_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, k:k_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data



# -*- coding: UTF-8 -*-

from typing import Any

from matplotlib_venn import venn3, venn3_circles, venn2, venn2_circles

from .. import util as ul
from ..util import path, collection, type_set_colors, plot_end, plot_start

__name__: str = "plot_venn"


def three_venn(
    set1: collection,
    set2: collection,
    set3: collection,
    name1: str = "Set1",
    name2: str = "Set2",
    name3: str = "Set3",
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    colors: list = None,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    fig, ax = plot_start(width, height, bottom, output, show)

    if colors is None:
        colors = type_set_colors[:3]

    if len(colors) < 3:
        ul.log(__name__).info(f"The value of colors requires three elements.")
        raise ValueError(f"The value of colors requires three elements.")
    elif len(colors) > 3:
        colors = colors[:3]

    set1 = set(set1)
    set2 = set(set2)
    set3 = set(set3)

    subsets = (set1, set2, set3)

    venn3(subsets=subsets, set_labels=(name1, name2, name3), ax=ax, set_colors=colors, **kwargs)

    # noinspection PyTypeChecker
    venn3_circles(subsets=subsets, linestyle='dashed', linewidth=1, color="grey", ax=ax)

    ax.legend(loc='upper right')

    ax.axis('off')

    plot_end(fig, title, x_name, y_name, output, show, close)


def two_venn(
    set1: collection,
    set2: collection,
    name1: str = "Set1",
    name2: str = "Set2",
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    colors: list = None,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    fig, ax = plot_start(width, height, bottom, output, show)

    if colors is None:
        colors = type_set_colors[:2]

    if len(colors) < 2:
        ul.log(__name__).info(f"The value of colors requires three elements.")
        raise ValueError(f"The value of colors requires three elements.")
    elif len(colors) > 2:
        colors = colors[:2]

    set1 = set(set1)
    set2 = set(set2)

    venn2((set1, set2), set_labels=(name1, name2), ax=ax, set_colors=colors, **kwargs)

    # noinspection PyTypeChecker
    venn2_circles(subsets=(set1, set2), linestyle='dashed', linewidth=1, color="grey", ax=ax)

    ax.legend(loc='upper right')

    ax.axis('off')

    plot_end(fig, title, x_name, y_name, output, show, close)
